#!/usr/bin/perl
#$rcs = ' $Id: indexer.pl,v 1.6 2000/08/01 21:01:50 daniel Exp $ ' ;	

# Perlfect Search
#
# Copyright (C) 1999-200 Giorgos Zervas <giorgos@perlfect.com>
# http extension by Daniel Naber <dnaber@mini.gt.owl.de>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

use Fcntl;

# added program path to @INC because it fails to find ./conf.pl if
# started from other directory
{ 
  # block is for $1 not mantaining its value
  $0 =~ /(.*)\//;
  push @INC, $1 if $1;
}

my $db_package = "";
package AnyDBM_File;
@ISA = qw(DB_File GDBM_File SDBM_File ODBM_File NDBM_File) unless @ISA;
foreach my $isa (@ISA) {
  if( eval("require $isa") ) {
    print "Using $isa...\n";
    $db_package = $isa;
    last;
  }
}

package main;
require 'conf.pl';
require 'indexer_web.pl';
require 'SGMLEntities.pl';

$|=1;

my $DN  = 0;  #documents number
my $TN  = 0;  #terms number

my @stopwords;
my @no_index;
my $special_chars = "";		# special characters to replace

my %inv_index_db;
my %docs_db;
my %desc_db;
my %title_db;
my %terms_db;
my %df_db;
my %tf_db;

tie %inv_index_db, $db_package, $INV_INDEX_TMP_DB_FILE, O_CREAT|O_RDWR, 0755 or die "Cannot open $INV_INDEX_TMP_DB_FILE: $!";   
tie %docs_db,      $db_package, $DOCS_TMP_DB_FILE, O_CREAT|O_RDWR, 0755      or die "Cannot open $DOCS_TMP_DB_FILE: $!";   
tie %desc_db,      $db_package, $DESC_TMP_DB_FILE, O_CREAT|O_RDWR, 0755      or die "Cannot open $DESC_TMP_DB_FILE: $!";   
tie %titles_db,    $db_package, $TITLES_TMP_DB_FILE, O_CREAT|O_RDWR, 0755    or die "Cannot open $TITLES_TMP_DB_FILE: $!";   
tie %terms_db,     $db_package, $TERMS_TMP_DB_FILE, O_CREAT|O_RDWR, 0755     or die "Cannot open $TERMS_TMP_DB_FILE: $!";   
tie %df_db,        $db_package, $DF_DB_FILE, O_CREAT|O_RDWR, 0755            or die "Cannot open $DF_DB_FILE: $!";
tie %tf_db,        $db_package, $TF_DB_FILE, O_CREAT|O_RDWR, 0755            or die "Cannot open $TF_DB_FILE: $!";

print "Loading \'no index\' regular expressions:\n";
load_excludes();
print "Loading stopwords...";
my $stopwords_regex = load_stopwords();
print "Done.\n";

init();
print "Starting crawler...\n";
# build list of special characters that will be replaced in normalize():
foreach my $number (keys %entities) {
  $special_chars .= chr($number);
}  
crawl($START_URL);
print "Crawler finished($DN files, $TN terms)\n\n";

print "Calculating weight vectors: \n";
weights();

print "Removing unused db files:\n";
# Removing the files is a problem on Windows NT if they are still tie'd:
untie %inv_index_db;
untie %docs_db;
untie %desc_db;
untie %titles_db;
untie %terms_db;
untie %df_db;
untie %tf_db;
print "\t$TF_DB_FILE...";
if (unlink $TF_DB_FILE) {
  print "ok\n";
} else {
  warn "Cannot unlink $TF_DB_FILE: $!";
}

print "\t$DF_DB_FILE...";
if (unlink $DF_DB_FILE) {
  print "ok\n";
} else {
  warn "Cannot unlink $DF_DB_FILE: $!";
}

print "Renaming newly created db files...\n";
rename_db();

print "Indexer finished.\n";

# http extension: removed crawl, it's now in indexer_web.pl

sub index_file {
  my ($url, $doc_id, $buffer) = @_;
  my ($term, $term_id);
  my %tf;

  # http extension: removed loading of file, now done in indexer_web.pl
  print "\t $doc_id: $url\n";

  record_desc($doc_id, $buffer);    
  if( $INDEX_URLS ) {
    # to search for parts of URLs, e.g. filenames:
    # (cutting out $DOCUMENT_ROOT for security reasons!)
    my ($file_part) = ($file =~ m/^$DOCUMENT_ROOT(.*)$/);
    $buffer = $file_part." ".$buffer;
  }
  # to search for text in the follwing meta tags:
  $buffer = get_meta_content("description", $buffer, $META_WEIGHT)." ".$buffer;
  $buffer = get_meta_content("keywords", $buffer, $META_WEIGHT)." ".$buffer;
  $buffer = get_meta_content("author", $buffer, $META_WEIGHT)." ".$buffer;
  # to search for images' alt texts:
  $buffer = get_alt_texts($buffer)." ".$buffer;
  $buffer = normalize($buffer);
  
  foreach (split " ", $buffer) {
    $_ = substr $_, 0, $STEMCHARS if $STEMCHARS;
    if (length $_ >= $MINLENGTH) {
      $term_id = record_term($_);
      ++$tf{$term_id};
    }
  }
  
  foreach (keys %tf) {
    $df_db{$_}++;
    $tf_db{$_} = '' unless defined $tf_db{$_};
    $tf_db{$_} .= pack("ww", $doc_id, $tf{$_}); 
  }
}

sub weights {
  my ($weight, $term_id, $doc_id);
  my $step = $TN / 50;
  my $count = 0;
  
  $step = 1 if $step < 1;

  print "0%  10%  20%  30%  40%  50%  60%  70%  80%  90%  100%\n";
  print "|----|----|----|----|----|----|----|----|----|----|\n";
  print ">";


  foreach $term_id (keys %tf_db) {
    my $weights = $inv_index_db{$term_id} || '';
    my $df = $df_db{$term_id};
    my %tdf = unpack("w*",$tf_db{$term_id}); 
    $count++;
    print "\b >" unless $count % $step;
    foreach $doc_id (keys %tdf) {
      #print "weight = $tdf{$doc_id} * log ($DN / $df\n";
      $weight = $tdf{$doc_id} * log ($DN / $df);
      $weights .= pack("ff", $doc_id, $weight);
    }
    undef %tdf;
    $inv_index_db{$term_id} = $weights;
  }
  print "\n";
}

sub normalize {
  my $buffer = $_[0];

  $buffer =~ s/<!--.*?-->//gis;  # strip html comments
  $buffer =~ s/-(\s*\n\s*)?//g;  # join parts of hyphenated words

  if( $SPECIAL_CHARACTERS ) {
    # There may be special characters that are not encoded, so encode them:
    $buffer =~ s/([$special_chars])/"&#".ord($1).";"/gse;
    # Special characters can be encoded using hex values:
    $buffer =~ s/&#x([\dA-F]{2});/"&#".hex("0x".$1).";"/igse;
    # Special charaters may be encoded with numbers, undo that:
    $buffer =~ s/&#(\d\d\d);/&$entities{$1};/gs;
    # Special cases:
    $buffer =~ s/&thorn;/th/igs;
    $buffer =~ s/&eth;/d/igs;
    $buffer =~ s/&szlig;/ss/igs;
    # Now represent special characters as the characters they are based on:
    $buffer =~ s/&(..?)(grave|acute|circ|tilde|uml|ring|cedil|slash|lig);/$1/igs;
  }

  if ($INDEX_NUMBERS) {
    $buffer =~ s/(<[^>]*>)/ /gs;
  } else {
    $buffer =~ s/(\b\d+\b)|(<[^>]*>)/ /gs;
  }

  $buffer =~ s/$stopwords_regex//gio;
  $buffer =~ tr/a-zA-Z0-9_/ /cs;
  return lc $buffer;
}

sub record_file {
  my $file = $_[0];

  $file =~ m/^$DOCUMENT_ROOT(.*)$/;
  $file = $1;
  unless ($file =~ /^\//) {
    $file = "/".$file;
  }

  ++$DN;
  $docs_db{$DN} = $file;
  return $DN;
}

sub record_desc {
  my ($doc_id, $buffer) = @_;
  my ($desc, $title);
  my @desc_ary;

  $desc = get_meta_content("description", $buffer, 1);
  
  unless ($desc) {
    $buffer =~ m/<BODY.*?>(.*)<\/BODY>/is;
    $desc = $1;
    $desc =~ s/<!--.*?-->//gis;  # strip html comments
    $desc =~ s/<.+?>/ /gis;      # strip html
    $desc =~ tr/\n\r/ /s;
    @desc_ary = split " ", $desc;
    $desc = $1;
    $desc = join " ", @desc_ary[0..$DESC_WORDS];
  }
  $desc_db{$doc_id} = $desc;

  $buffer =~ m/<TITLE>(.*?)<\/TITLE>/is;
  $title = $1;
  $title ||= "Untitled";
  $titles_db{$doc_id} = $title;
}

# Get the content part for a certain meta tag. Weight with
# a certain factor by just repeating the result that often:
sub get_meta_content {
  my $name = $_[0];
  my $buffer = $_[1];
  my $weight = $_[2];
  my ($content) = ($buffer =~ m/<META\s+name\s*=\s*[\"\']?$name[\"\']? content=[\"\']?(.*?)[\"\']?>/is);
  $content = (($content." ") x $weight);
  return if( ! $content || $content =~ m/^\s+$/ );
  return $content;  
}

# Get all values for alt="...", joined with spaces:
sub get_alt_texts {
  my $buffer = $_[0];
  my $alt_texts = "";
  while( $buffer =~ m/alt=[\"\'](.*?)[\"\']/gis ) {
  	$alt_texts .= " ".$1;
  }
  return $alt_texts;
}

sub record_term {
  my $term = $_[0];

  print STDERR "Warning: record_term($term): No term was supplied\n" unless $term;

  if ($terms_db{$term}) {
    return $terms_db{$term};
  } else {
    ++$TN;
    $terms_db{$term} = $TN;
    return $TN;
  }
}

sub load_excludes {
  if (-e $NO_INDEX_FILE) {
    open (FILE, $NO_INDEX_FILE) or (warn "Cannot open $NO_INDEX_FILE: $!" and next);
    while (<FILE>) {
      chomp;
      $_ =~ s/\r//g;        # get rid of carriage returns      
      $_ =~ s/(\#.*)//g;    # ingore comments
      $_ =~ s/[\/\s]*$//;   # remove any trailing spaces and slashes
      print "\t- $_\n";
      
      $_ = quotemeta;       # escape all non-alphanumeric characters
      $_ =~ s/\\\*/\.\*/g;  # except for the * which is replaced by .*
      push @no_index, $_;
    }
    close (FILE);
  } else {
    print STDERR "Warning: $NO_INDEX_FILE missing.";
  }
}

sub rename_db {
  my @files = (
	       [$TERMS_TMP_DB_FILE, $TERMS_DB_FILE],
	       [$DOCS_TMP_DB_FILE, $DOCS_DB_FILE],
	       [$TITLES_TMP_DB_FILE, $TITLES_DB_FILE],
	       [$DESC_TMP_DB_FILE, $DESC_DB_FILE],
	       [$INV_INDEX_TMP_DB_FILE, $INV_INDEX_DB_FILE],
	      );

  foreach (@files) {
    if (-e $_->[0]) {
      print "\t ", $_->[0], " to ", $_->[1], "\n";
      rename $_->[0], $_->[1] or (warn "Cannot rename $_: $!" and next);
    } else {
      warn "File ", $_->[0], " doesn't exist: $!" and next;
    }
  }
}

sub load_stopwords {
  my @stopwords;
  my $stopwords_regex;

  open (FILE, $STOPWORDS_FILE) or (warn "Cannot open $STOPWORDS_FILE: $!" and next);
  while (<FILE>) {
    chomp;
    $_ =~ s/\r//g; # get rid of carriage returns
    push @stopwords, $_;
  }
  close(FILE);
  $stopwords_regex = '\b(' . join('|', @stopwords) . ')\b';

  return $stopwords_regex;
}
